#include <clip.h>

#define SCALE		2

#define ROM(fname)	"C:\\Home\\Emulator\\colecovision\\roms\\" fname
//#define BIOS		ROM("ColecoVision BIOS (1982).col")
#define BIOS		ROM("ColecoVision BIOS (1982) [h1] (no title delay).col")
//#define GAME		ROM("Mr. Do's Castle (1983) (Parker Bros).col")
#define GAME		ROM("Donkey Kong (1982).col")
//#define GAME		ROM("Alcazar - The Forgotten Fortress (1985) (Activision).col")
//#define GAME		ROM("Boulder Dash (1984) (Micro Fun).col")

typedef struct _DRIVER {
	Z80 cpu;
	unsigned char mem[0x10000];
	//
	TMS9918 vdp;
} DRIVER;
DRIVER _driver;
DRIVER* driver = &_driver;

int joymode;

unsigned char
z80_in(Z80* z80, unsigned short addr)
{
	int data;

	addr &= 0xff;
	switch(addr & 0xe0) {
	//case 0x40: /* v^Xe[^X (ADAMp) */
	case 0xa0: /* TMS9918A */
		return tms9918_read(&driver->vdp, addr & 1);
	case 0xe0: /* L[pbhAWCXeBbN (0xe0=Player1, 0xe1=Player2) */
		if(addr & 1) break;

#define JOYPAD_UP	(~(1<<0))
#define JOYPAD_RI	(~(1<<1))
#define JOYPAD_DN	(~(1<<2))
#define JOYPAD_LF	(~(1<<3))
#define JOYPAD_FIRE2	(~(1<<6))
#define KEYPAD_0	0xfa
#define KEYPAD_1	0xfd
#define KEYPAD_2	0xf7
#define KEYPAD_3	0xfc
#define KEYPAD_4	0xf2
#define KEYPAD_5	0xf3
#define KEYPAD_6	0xfe
#define KEYPAD_7	0xf5
#define KEYPAD_8	0xf1
#define KEYPAD_9	0xfb
//#define KEYPAD_STER	0xf6
//#define KEYPAD_POUND	0xf9
//ColEm(The Portable ColecoVision Emulator)VColeco(The Virtual ColecoVision)́A
//'*''#'t̂悤ȋC܂BMEKA(Multi machine emulator)݂͐łB
#define KEYPAD_STER	0xf9
#define KEYPAD_POUND	0xf6
#define JOYPAD_FIRE1	(~(1<<6))

		data = 0xff;
		if(joymode) { /* WCXeBbN */
			/* bit0: Up    (0=On,1=Off)
			 * bit1: Right (0=On,1=Off)
			 * bit2: Down  (0=On,1=Off)
			 * bit3: Left  (0=On,1=Off)
			 * bit6: Fire2 (0=On,1=Off)
			 */
			if(GetAsyncKeyState(VK_UP   ) & 0x8000) data &= JOYPAD_UP;
			if(GetAsyncKeyState(VK_RIGHT) & 0x8000) data &= JOYPAD_RI;
			if(GetAsyncKeyState(VK_DOWN ) & 0x8000) data &= JOYPAD_DN;
			if(GetAsyncKeyState(VK_LEFT ) & 0x8000) data &= JOYPAD_LF;
			if(GetAsyncKeyState('X'     ) & 0x8000) data &= JOYPAD_FIRE2;

		} else { /* L[pbh */
			/* bit3-0: 15=Off
			 * +-----+  0='0'  6='6'
			 * |1 2 3|  1='1'  7='7'
			 * |4 5 6|  2='2'  8='8'
			 * |7 8 9|  3='3'  9='9'
			 * |* 0 #|  4='4' 10='*'
			 * +-----+  5='5' 11='#'
			 */
			if(GetAsyncKeyState('0') & 0x8000) data = KEYPAD_0;
			if(GetAsyncKeyState('1') & 0x8000) data = KEYPAD_1;
			if(GetAsyncKeyState('2') & 0x8000) data = KEYPAD_2;
			if(GetAsyncKeyState('3') & 0x8000) data = KEYPAD_3;
			if(GetAsyncKeyState('4') & 0x8000) data = KEYPAD_4;
			if(GetAsyncKeyState('5') & 0x8000) data = KEYPAD_5;
			if(GetAsyncKeyState('6') & 0x8000) data = KEYPAD_6;
			if(GetAsyncKeyState('7') & 0x8000) data = KEYPAD_7;
			if(GetAsyncKeyState('8') & 0x8000) data = KEYPAD_8;
			if(GetAsyncKeyState('9') & 0x8000) data = KEYPAD_9;
			if(GetAsyncKeyState('A') & 0x8000) data = KEYPAD_STER; /*'*'*/
			if(GetAsyncKeyState('S') & 0x8000) data = KEYPAD_POUND; /*'#'*/
			if(GetAsyncKeyState('Z') & 0x8000) data &= JOYPAD_FIRE1;
		}
		return data;
	}

	return 0xff;
}

void
z80_out(Z80* z80, unsigned short addr, unsigned char data)
{
	addr &= 0xff;
	switch(addr & 0xe0) {
	//case 0x40: /* v^o (ADAMp) */
	case 0x80: /* L[pbhǂݍݑI */
		joymode = 0;
		break;
	case 0xa0: /* TMS9918A (0xa0=VRAM, 0xa1=Register) */
		tms9918_write(&driver->vdp, addr & 1, data);
		break;
	case 0xc0: /* WCXeBbNǂݍݑI */
		joymode = 1;
		break;
	//case 0xe0: /* SN76489 */
	}
}

unsigned char xbuff[256 * 192];

BITMAPINFO* bmi;

void
init()
{
	BITMAPINFOHEADER* bi;
	RGBQUAD* rgb;
	int i;

	bmi = (BITMAPINFO*)calloc(sizeof(BITMAPINFOHEADER) + sizeof(RGBQUAD) * 256, 1);
	bi = &bmi->bmiHeader;
	rgb = &bmi->bmiColors[0];
	bi->biSize = sizeof(BITMAPINFOHEADER);
#ifdef TMS9918_GRAYSCALE
	bi->biWidth = 128;
	bi->biHeight = -88;
#else 
	bi->biWidth = 256;
	bi->biHeight = -192;
#endif
	bi->biPlanes = 1;
	bi->biBitCount = 8;
	for(i = 0; i < 16; i++) {
#ifdef TMS9918_GRAYSCALE
		rgb[i].rgbRed   = 255 * i / 15 ^ 255;
		rgb[i].rgbGreen = 255 * i / 15 ^ 255;
		rgb[i].rgbBlue  = 255 * i / 15 ^ 255;
#else
		rgb[i].rgbRed   = tms9918_color_table[i][0];
		rgb[i].rgbGreen = tms9918_color_table[i][1];
		rgb[i].rgbBlue  = tms9918_color_table[i][2];
#endif
	}
}

int
main(int argc, char* argv[])
{
	const char* fname;
	FILE* fp;
	HDC hdc;
	int t0;
	int t1;

	init();

	memset(driver->mem, -1, 0x10000);

	fp = fopen(BIOS, "rb");
	if(!fp) DIE();
	fread(&driver->mem[0x0000], 1, 0x2000, fp); /* just 8KB */
	fclose(fp);

	fname = argv[1] ? argv[1] : GAME;
	fp = fopen(fname, "rb");
	fread(&driver->mem[0x8000], 1, 0x8000, fp); /* max 32KB */
	fclose(fp);

	tms9918_reset(&driver->vdp);
	z80_reset(&driver->cpu, NULL, NULL, z80_in, z80_out);

	t0 = GetTickCount();
	for(;;) {
		z80_run(&driver->cpu, NTSC_COLOR_SUBCARRIER / 60);
		if(tms9918_update(&driver->vdp, xbuff) & 0x80) {
			z80_nmi(&driver->cpu);
		}
#ifdef TMS9918_GRAYSCALE
		tms9918_reduce(xbuff, xbuff);
#endif

		hdc = GetDC(NULL);
		StretchDIBits(hdc,
			0, 0, 256 * SCALE, 192 * SCALE,
			0, 0, bmi->bmiHeader.biWidth, -bmi->bmiHeader.biHeight,
			xbuff,
			bmi,
			DIB_RGB_COLORS,
			SRCCOPY);
		ReleaseDC(NULL, hdc);

		for(;;) {
			t1 = GetTickCount();
			if(t1 - t0 >= 1000 / 60) break;
			Sleep(0);
		}
		t0 += 1000 / 60;

		if(GetAsyncKeyState(VK_F8) & 0x8000) exit(0);
	}
}

