/*
 *	emu.c
 *
 *	ColecoVision/EMU
 *
 *	* Sun Jan 23 00:00:00 JST 2005 Naoyuki Sawa
 *	- 1st [XB
 */
#include "app.h"

/****************************************************************************
 *	
 ****************************************************************************/

#define JOYSTICK_UP	(~(1<<0))
#define JOYSTICK_RI	(~(1<<1))
#define JOYSTICK_DN	(~(1<<2))
#define JOYSTICK_LF	(~(1<<3))
#define JOYSTICK_FIRE2	(~(1<<6))
#define KEYPAD_0	0xfa
#define KEYPAD_1	0xfd
#define KEYPAD_2	0xf7
#define KEYPAD_3	0xfc
#define KEYPAD_4	0xf2
#define KEYPAD_5	0xf3
#define KEYPAD_6	0xfe
#define KEYPAD_7	0xf5
#define KEYPAD_8	0xf1
#define KEYPAD_9	0xfb
//#define KEYPAD_STER	0xf6
//#define KEYPAD_POUND	0xf9
//ColEm(The Portable ColecoVision Emulator)VColeco(The Virtual ColecoVision)́A
//'*''#'t̂悤ȋC܂BMEKA(Multi machine emulator)݂͐łB
#define KEYPAD_STER	0xf9
#define KEYPAD_POUND	0xf6
#define JOYSTICK_FIRE1	(~(1<<6))

int
coleco_in(int addr)
{
	switch(addr & 0xe0) {

	//case 0x40: /* v^Xe[^X (ADAMp) */

	case 0xa0: /* TMS9918A */
		return tms9918_read(&coleco_vdp, addr & 1);

	case 0xe0: /* L[pbhAWCXeBbN (0xe0=Player1, 0xe1=Player2) */
		if(addr & 1) { /* Player2 */
			return -1;
		}
		return joystat[joymode];
	}

	return -1;
}

void
coleco_out(int addr, int data)
{
	switch(addr & 0xe0) {

	//case 0x40: /* v^o (ADAMp) */

	case 0x80: /* L[pbhǂݍݑI */
		joymode = 0;
		return;

	case 0xa0: /* TMS9918A (0xa0=VRAM, 0xa1=Register) */
		tms9918_write(&coleco_vdp, addr & 1, data);
		return;

	case 0xc0: /* WCXeBbNǂݍݑI */
		joymode = 1;
		return;

	case 0xe0: /* SN76489 */
		ENTER_CS; /*Kv!!*/
		sn76489_write(&coleco_dcsg, data);
		LEAVE_CS; /*Kv!!*/
		return;
	}
}

/****************************************************************************
 *	
 ****************************************************************************/

#define KEYPAD_X	8
#define KEYPAD_Y	8

static const unsigned char keymap[4][3] = {
	{KEYPAD_1,   KEYPAD_2,KEYPAD_3    },
	{KEYPAD_4,   KEYPAD_5,KEYPAD_6    },
	{KEYPAD_7,   KEYPAD_8,KEYPAD_9    },
	{KEYPAD_STER,KEYPAD_0,KEYPAD_POUND},
};

void
keypad_disp(int blink)
{
	render_object(&render, KEYPAD_X, KEYPAD_Y, 0, 0, 31, 72, DRW_NOMAL);
	if(blink) {
		render_string_framed(&render, KEYPAD_X + 38, KEYPAD_Y + 42, "FI", 0x80, 0x00f);
		render_string_framed(&render, KEYPAD_X + 38, KEYPAD_Y + 52, "`a F", 0x80, 0x00f);
		render_string_framed(&render, KEYPAD_X + 38, KEYPAD_Y + 62, "STSELF", 0x80, 0x00f);
	}
	if(!blink || (now % SEC(0.5) < SEC(0.3))) {
		render_rectangle(&render,
			(KEYPAD_X +  2) + keypad_col * 9,
			(KEYPAD_Y + 34) + keypad_row * 9, 9, 9, 0);
	}
}

void
keypad()
{
	pceWaveSetChAtt(0, 127);

	keystat = -1;
	for(;;) {
		schedule();

		if(joy & TRG_CD) {
			break;
		}
		if(joy & TRG_AB) {
			keystat = keymap[keypad_row][keypad_col];
			break;
		}
		if((joy & TRG_LF) && (keypad_col > 0)) keypad_col--;
		if((joy & TRG_RI) && (keypad_col < 2)) keypad_col++;
		if((joy & TRG_UP) && (keypad_row > 0)) keypad_row--;
		if((joy & TRG_DN) && (keypad_row < 3)) keypad_row++;

		keypad_disp(1);
	}

	pceWaveSetChAtt(0, 0);
}

/****************************************************************************
 *	
 ****************************************************************************/

int
coleco_stream_callback(short* wbuff, int param)
{
	sn76489_mix(&coleco_dcsg, wbuff, 320);
	return 0;
}

int
coleco_run(const char* fname, int sound_on)
{
	int retval;
	int cnt;
	FILE* fp;

	/********************************************************************
	 *	
	 ********************************************************************/

	memset(z80mem, -1, 0x10000);

	joymode    =  0;
	joystat[0] = -1;
	joystat[1] = -1;
	keystat    = -1;
	keypad_col =  0;
	keypad_row =  0;

	/* BIOS[hB */
	fp = fopen(bios_fname, "rb");
	if(!fp) {
		return -1;
	}
	fread(&z80mem[0x0000], 1, 0x2000, fp); /* 8KB */
	fclose(fp);

	/* ROM[hB */
	fp = fopen(fname, "rb");
	if(!fp) {
		return -1;
	}
	fread(&z80mem[0x8000], 1, 0x8000, fp); /* MAX 32KB */
	fclose(fp);

	z80cpu_reset(coleco_in, coleco_out);
	tms9918_reset(&coleco_vdp);
	sn76489_reset(&coleco_dcsg, NTSC_COLOR_SUBCARRIER);
	if(sound_on) {
		stream_play(320, coleco_stream_callback, 0, 0);
	}

	/********************************************************************
	 *	s
	 ********************************************************************/

	cnt = 0;
	for(;;) {
		z80cpu_run(NTSC_COLOR_SUBCARRIER / 60);
		cnt += FPS;
		if(cnt >= 60) {
			cnt -= 60;

			/* \XVB */
			retval = tms9918_update(&coleco_vdp, vbuff);
			tms9918_reduce(vbuff, vbuff);
			if(keystat != -1) {
				keypad_disp(0); /* L[pbh͒́AL[pbh\ێ */
			}
			schedule();

			/* ͍XV */
			if(joy & TRG_SELECT) {
				break;
			}
			if(joy & TRG_START) {
				keypad(); /* L[pbh */
			}
			joystat[0] = -1;
			joystat[1] = -1;
			if(keystat != -1) {
				if(joy & PAD_AB) {
					joystat[0] = keystat; /* L[pbh͒ */
				} else {
					keystat = -1; /* L[pbh͉ */
				}
			} else {
				if(joy & PAD_UP) joystat[1] &= JOYSTICK_UP;
				if(joy & PAD_DN) joystat[1] &= JOYSTICK_DN;
				if(joy & PAD_LF) joystat[1] &= JOYSTICK_LF;
				if(joy & PAD_RI) joystat[1] &= JOYSTICK_RI;
				if(joy & PAD_A ) joystat[1] &= JOYSTICK_FIRE2;
				if(joy & PAD_B ) joystat[0] &= JOYSTICK_FIRE1;
			}

		} else {
			retval = tms9918_update(&coleco_vdp, NULL);
		}

		if(retval & 0x80/*Interrupt Flag*/) {
			z80cpu_nmi();
		}
	}

	/********************************************************************
	 *	N[Abv
	 ********************************************************************/

	if(sound_on) {
		stream_stop();
	}

	return 0;
}

